#include <windows.h>
#include <bwcc.h>
#include <string.h>
#include <print.h>
#include "choice.h"
#include "ovprint.h"



extern HANDLE hInstDLL;


#define MAXPORTS	11
#define NAMESIZE	40
#define DRIVERSIZE	14
#define PORTSIZE        9       // must be large enough for LPT1.OS2

#define CHANGEFIELDS DM_ORIENTATION|DM_PAPERSIZE|DM_COPIES|DM_DEFAULTSOURCE|DM_PRINTQUALITY|DM_COLOR|DM_DUPLEX|DM_YRESOLUTION|DM_TTOPTION


struct DriverPort{
  char name[NAMESIZE];
  char driver[DRIVERSIZE];
  char port[PORTSIZE];
} DPArray[MAXPORTS];

#define BUFSIZE		sizeof(DPArray) + 3 * MAXPORTS

char IniFile[81];
char Application[31];

char  OldPrinter[81];
char  NewPrinter[81];
char  NewName[NAMESIZE];
char  NewDriver[DRIVERSIZE];
char  NewPort[PORTSIZE];
short OldOrientation;
short OldPaperSize;
short OldCopies;
short OldDefaultSource;
short OldPrintQuality;
short OldColor;
short OldDuplex;
short OldYResolution;
short OldTTOption;
short NewOrientation;
short NewPaperSize;
short NewCopies;
short NewDefaultSource;
short NewPrintQuality;
short NewColor;
short NewDuplex;
short NewYResolution;
short NewTTOption;

/************************************************
Message Functions
************************************************/
void MsgNoLoad(HANDLE hWindow)
{
   char Buffer[80];
   wsprintf(Buffer, "Could not load %s", NewDriver);
   BWCCMessageBox(hWindow, Buffer, NULL, MB_OK);
}

void MsgNoExtDeviceMode(HANDLE hWindow)
{
   char Buffer[80];
   wsprintf(Buffer, "%s doesn't support ExtDeviveMode", NewDriver);
   BWCCMessageBox(hWindow, Buffer, NULL, MB_OK);

}

void MsgNotValidSetting(HANDLE hWindow, LPSTR pinifile, LPSTR papplication)
{
   char Buffer[80];
   wsprintf(Buffer, "%s [%s] not a valid setting", pinifile, papplication);
   BWCCMessageBox(hWindow, Buffer, NULL, MB_OK);
}

void MsgIniWriteError(HANDLE hWindow, LPSTR pinifile)
{
   char Buffer[80];
   wsprintf(Buffer, "Error writing to %s", pinifile);
   BWCCMessageBox(hWindow, Buffer, NULL, MB_OK);
}

/************************************************
non exported function to reset current NewDriver
************************************************/
void ResetNewDriver(HANDLE hWindow)
{
   LPDEVMODE lpDevMode;
   HANDLE hDevMode;
   HANDLE hPrintDrv;
   int size;
   LPFNDEVMODE lpfnExtDeviceMode;
   char Buffer[80];

   //  only proceed if name, driver & port are non null stings
   if (*NewName) {

     // get the new driver's current settings
     hPrintDrv = LoadLibrary(NewDriver);
     if (hPrintDrv >= HINSTANCE_ERROR) {
        lpfnExtDeviceMode = (LPFNDEVMODE) GetProcAddress(hPrintDrv,
                                                      "ExtDeviceMode");
        if (lpfnExtDeviceMode) {

	   size = (*lpfnExtDeviceMode)(NULL, hPrintDrv, NULL,
				    NewName,
				    NewPort,
				    NULL, NULL, 0);

	   hDevMode = GlobalAlloc(GMEM_FIXED|GMEM_ZEROINIT, size);
	   lpDevMode = (LPDEVMODE)GlobalLock(hDevMode);

	   (*lpfnExtDeviceMode)(hWindow, hPrintDrv, lpDevMode,
				    NewName,
				    NewPort,
				    NULL,
				    NULL, DM_COPY);

	   // reset the printer driver
	   lpDevMode->dmOrientation = OldOrientation;
	   lpDevMode->dmPaperSize = OldPaperSize;
           lpDevMode->dmCopies = OldCopies;
	   lpDevMode->dmDefaultSource = OldDefaultSource;
	   lpDevMode->dmPrintQuality = OldPrintQuality;
           lpDevMode->dmColor = OldColor;
	   lpDevMode->dmDuplex = OldDuplex;
	   lpDevMode->dmYResolution = OldYResolution;
           lpDevMode->dmTTOption = OldTTOption;

	   //no need to update fields that aren't altered
           lpDevMode->dmFields = CHANGEFIELDS;

	   // don't need to worry about the trailing ":" being
	   // stripped from NewPortsince this code section can't
	   // be reached again until after it is reset by SetPrinter
	   (*lpfnExtDeviceMode)(hWindow, hPrintDrv, NULL,
				    NewName,
				    NewPort,
				    lpDevMode,
				    NULL, DM_MODIFY|DM_UPDATE);

	  GlobalUnlock(hDevMode);
	  GlobalFree(hDevMode);
	}
	else {
	  MsgNoExtDeviceMode(hWindow);
	}
	FreeLibrary(hPrintDrv);
     }
     else {
	MsgNoLoad(hWindow);
     }
   }
}


/***********************************************
Dialog function for SetupPrinter's dialog box
Fills list box with available printer/port selections.
If printer is selected & OK clicked, printer driver's
setup dialog is called.
If printer diriver's dialog OK is clicked, settings
are saved in .ini file
************************************************/

// Turn off "Parameter never used" warning
#pragma argsused

BOOL FAR PASCAL SelPrnDlgProc(HWND hDlg, WORD message, WORD wParam,
                              LONG lParam)
{
  HANDLE hBuffer1, hBuffer2, hDevModeOut;
  LPSTR pBuffer1,pBuffer2,p1,p2;
  int i, x;
  DWORD ret;
  char tempname[NAMESIZE];
  char tempdriver[DRIVERSIZE];
  char *pport;

  int size;
  LPDEVMODE lpDevModeOut;
  HANDLE hPrintDrv;
  LPFNDEVMODE lpfnExtDeviceMode;

  switch (message) {
    case WM_INITDIALOG:

      // create buffers
      hBuffer1 = GlobalAlloc(GMEM_FIXED|GMEM_ZEROINIT, BUFSIZE);
      pBuffer1 = GlobalLock(hBuffer1);
      hBuffer2 = GlobalAlloc(GMEM_FIXED|GMEM_ZEROINIT, 81);
      pBuffer2 = GlobalLock(hBuffer2);

      // get all the device entries
      GetProfileString("devices", NULL, "", pBuffer1, BUFSIZE-1);
      p1 = pBuffer1;

      // get driver/port information
      for (i=0;*p1;++p1) {

	 // temporarily hold name and device information
	 lstrcpy(tempname, p1);
	 //get the profile string
	 GetProfileString("devices", tempname, "", pBuffer2, 80);
	 p2 = pBuffer2;
	 lstrcpy(tempdriver, _fstrtok(p2, ","));

	 // fill in struct for each port for which driver is available
	 while ((pport = _fstrtok(NULL, ",")) != NULL && i < MAXPORTS){
	    lstrcpy(DPArray[i].name, tempname);
	    lstrcpy(DPArray[i].driver, tempdriver);
	    lstrcat(DPArray[i].driver, ".DRV");
	    lstrcpy(DPArray[i].port, pport);
	    ++i;
	 }
	 p1 = _fstrchr(p1,0);
      }

      // insert name/port sting into list box
      DPArray[i].name[0] = '\0';
      for (i=0; DPArray[i].name[0] != '\0'; ++i) {
	 wsprintf(pBuffer2, "%s on %s", DPArray[i].name, DPArray[i].port);
	 SendDlgItemMessage(hDlg, PRINTER_LB, LB_INSERTSTRING, -1,
			    (DWORD)pBuffer2);
      }

      // unlock and free buffers
      GlobalUnlock(hBuffer1);
      GlobalFree(hBuffer1);
      GlobalUnlock(hBuffer2);
      GlobalFree(hBuffer2);

      return TRUE;

    case WM_COMMAND:
      switch (wParam) {
	case IDOK:
	   // has a printer been selected?
	   ret = SendDlgItemMessage(hDlg, PRINTER_LB, LB_GETCURSEL,0,0);

	   // if a printer selected set it up
	   if (ret != LB_ERR) {

	     // load the driver module
	     hPrintDrv = LoadLibrary(DPArray[ret].driver);
             if (hPrintDrv >= HINSTANCE_ERROR){
	       lpfnExtDeviceMode = (LPFNDEVMODE)GetProcAddress(hPrintDrv,
	                                                    "ExtDeviceMode");
	       if (lpfnExtDeviceMode) {
	          size = (*lpfnExtDeviceMode)(hDlg, hPrintDrv, NULL,
					   DPArray[ret].name,
					   DPArray[ret].port,
					   NULL, NULL, 0);

	          hDevModeOut = GlobalAlloc(GMEM_FIXED|GMEM_ZEROINIT, size);
	          lpDevModeOut = (LPDEVMODE)GlobalLock(hDevModeOut);
	          x = (*lpfnExtDeviceMode)(hDlg, hPrintDrv, lpDevModeOut,
				    DPArray[ret].name,
				    DPArray[ret].port,
				    NULL,
				    NULL, DM_COPY|DM_PROMPT);

	          if ( x == IDOK) {
	             hBuffer2 = GlobalAlloc(GMEM_FIXED|GMEM_ZEROINIT, 6);
	             pBuffer2 = GlobalLock(hBuffer2);
	             if (WritePrivateProfileString(Application, "Name",
					 DPArray[ret].name,
					 IniFile)) {
		       WritePrivateProfileString(Application, "Driver",
					 DPArray[ret].driver,
					 IniFile);
		       WritePrivateProfileString(Application, "Port",
					 DPArray[ret].port,
					 IniFile);

		       wsprintf(pBuffer2,"%i", lpDevModeOut->dmOrientation);
		       WritePrivateProfileString(Application, "Orientation",
					 pBuffer2,
					 IniFile);

		       wsprintf(pBuffer2,"%i", lpDevModeOut->dmPaperSize);
		       WritePrivateProfileString(Application, "Size",
					 pBuffer2,
					 IniFile);

		       wsprintf(pBuffer2,"%i", lpDevModeOut->dmCopies);
		       WritePrivateProfileString(Application, "Copies",
					 pBuffer2,
					 IniFile);

		       wsprintf(pBuffer2,"%i", lpDevModeOut->dmDefaultSource);
		       WritePrivateProfileString(Application, "Source",
					 pBuffer2,
					 IniFile);

		       wsprintf(pBuffer2,"%i", lpDevModeOut->dmPrintQuality);
		       WritePrivateProfileString(Application, "PrintQuality",
					 pBuffer2,
					 IniFile);

		       wsprintf(pBuffer2,"%i", lpDevModeOut->dmColor);
		       WritePrivateProfileString(Application, "Color",
					 pBuffer2,
					 IniFile);

		       wsprintf(pBuffer2,"%i", lpDevModeOut->dmDuplex);
		       WritePrivateProfileString(Application, "Duplex",
					 pBuffer2,
					 IniFile);

		       wsprintf(pBuffer2,"%i", lpDevModeOut->dmYResolution);
		       WritePrivateProfileString(Application, "YResolution",
					 pBuffer2,
					 IniFile);

		       wsprintf(pBuffer2,"%i", lpDevModeOut->dmTTOption);
		       WritePrivateProfileString(Application, "TTOption",
					 pBuffer2,
					 IniFile);

		     }
	             else {
	               MsgIniWriteError(hDlg, IniFile);
		     }
	            GlobalUnlock(hBuffer2);
	            GlobalFree(hBuffer2);
	          }
	          GlobalUnlock(hDevModeOut);
	          GlobalFree(hDevModeOut);
	        }
	       else {
	          MsgNoExtDeviceMode(hDlg);
	       }
	       FreeLibrary(hPrintDrv);
	     }
	     else {
	       MsgNoLoad(hDlg);
	     }
 	     EndDialog(hDlg, 0);
	   }
	   // otherwise just display a message
	   else {
	      BWCCMessageBox(hDlg, "No printer selected.  Select an available printer or click Cancel to close.", "Message", MB_OK);
	   }
	   return TRUE;

	case IDCANCEL:
	  EndDialog(hDlg, 0);
	  return TRUE;
      }
  }
  return FALSE;
}

/********************************************
Setup printer and save settings in .ini file
********************************************/
// Argument template = "ACCyz"
BOOL FAR PASCAL _export SetupPrinter(LPSTR pIniFile, LPSTR pApplication,
				       HWND hMainWindow, HANDLE hInst)
{

  FARPROC DialogFunc;
  int ret;
  lstrcpy(IniFile, pIniFile);
  lstrcpy(Application, pApplication);
  DialogFunc = MakeProcInstance(SelPrnDlgProc, hInst);
  ret = DialogBox(hInstDLL, "SELECTPRINTERDLG", hMainWindow, DialogFunc);
  FreeProcInstance(DialogFunc);
  if (ret == -1) {
    MessageBox(hMainWindow, "Could not create dialog", NULL, MB_OK);
  }
  return TRUE;
}



/******************************************
Set the default printer to that specified
in an .ini file
******************************************/

// Argument template = "ACCy"

BOOL FAR PASCAL _export SetPrinter(LPSTR pIniFile,
				   LPSTR pApplication,
				   HANDLE hWindow)
{
   LPDEVMODE lpDevMode;
   HANDLE hDevMode;
   HANDLE hPrintDrv;
   int size;
   LPFNDEVMODE lpfnExtDeviceMode;
   char TempNewPort[PORTSIZE];      //work around for ":" stripping
   char NewDriverNoExt[DRIVERSIZE]; //NewDriver without ".DRV" 

   // first save the current default printer profile string
   // but only if SetPrinter hasn't been called yet
   // or has been called with subsequent call to ResetPrinter
   if (!*NewName) {
      GetProfileString("windows", "device", (LPSTR)NULL, OldPrinter,
                       sizeof(OldPrinter));
   }
   // If NewName is not a null string, then we are dealing with
   // back to back calls to SetPrinter
   // better reset the current NewDriver before proceeding
   else {
      ResetNewDriver(hWindow);
   }   
   // get custom printer settings from the ini file
   GetPrivateProfileString(pApplication, "Name", "", NewName,
			   sizeof(NewName), pIniFile);
   GetPrivateProfileString(pApplication, "Driver", "", NewDriver,
			   sizeof(NewDriver), pIniFile);
   GetPrivateProfileString(pApplication, "Port", "", NewPort,
			   sizeof(NewPort), pIniFile);
   NewOrientation = GetPrivateProfileInt(pApplication, "Orientation", 0,
					 pIniFile);
   NewPaperSize = GetPrivateProfileInt(pApplication, "Size", 0,
				       pIniFile);
   NewCopies = GetPrivateProfileInt(pApplication, "Copies", 0,
					 pIniFile);
   NewDefaultSource = GetPrivateProfileInt(pApplication, "Source", 0,
					   pIniFile);
   NewPrintQuality = GetPrivateProfileInt(pApplication, "PrintQuality", 0,
					 pIniFile);
   NewColor = GetPrivateProfileInt(pApplication, "Color", 0,
					 pIniFile);
   NewDuplex = GetPrivateProfileInt(pApplication, "Duplex", 0,
				    pIniFile);
   NewYResolution = GetPrivateProfileInt(pApplication, "YResolution", 0,
					 pIniFile);
   NewTTOption = GetPrivateProfileInt(pApplication, "TTOption", 0,
					 pIniFile);

   //  only proceed if name, driver & port are non null stings
   if (*NewName && *NewDriver && *NewPort) {

     // get the new driver's current settings and save the ones we change
     hPrintDrv = LoadLibrary(NewDriver);
     if (hPrintDrv >= HINSTANCE_ERROR) {
       lpfnExtDeviceMode = (LPFNDEVMODE) GetProcAddress(hPrintDrv,
                                                      "ExtDeviceMode");
       if (lpfnExtDeviceMode) {

	  size = (*lpfnExtDeviceMode)(NULL, hPrintDrv, NULL,
				    NewName,
				    NewPort,
				    NULL, NULL, 0);


	  hDevMode = GlobalAlloc(GMEM_FIXED|GMEM_ZEROINIT, size);
	  lpDevMode = (LPDEVMODE)GlobalLock(hDevMode);

	  (*lpfnExtDeviceMode)(hWindow, hPrintDrv, lpDevMode,
				    NewName,
				    NewPort,
				    NULL,
				    NULL, DM_COPY);
	  OldOrientation = lpDevMode->dmOrientation;
	  OldPaperSize = lpDevMode->dmPaperSize;
          OldCopies = lpDevMode->dmCopies;
	  OldDefaultSource = lpDevMode->dmDefaultSource;
	  OldPrintQuality = lpDevMode->dmPrintQuality;
          OldColor = lpDevMode->dmColor;
	  OldDuplex = lpDevMode->dmDuplex;
	  OldYResolution = lpDevMode->dmYResolution;
	  OldTTOption = lpDevMode->dmTTOption;


	  // now change the settings and update the driver
	  lpDevMode->dmOrientation = NewOrientation;
	  lpDevMode->dmPaperSize = NewPaperSize;
	  lpDevMode->dmCopies = NewCopies;
	  lpDevMode->dmDefaultSource = NewDefaultSource;
	  lpDevMode->dmPrintQuality = NewPrintQuality;
	  lpDevMode->dmColor = NewColor;
	  lpDevMode->dmDuplex = NewDuplex;
	  lpDevMode->dmYResolution = NewYResolution;
	  lpDevMode->dmTTOption = NewTTOption;

	  //tell the driver which fields to update
          lpDevMode->dmFields = CHANGEFIELDS;

	  // For some reason the HPLaserJet drivers and maybe others
	  // strip NewPort of a trailing ":" after next ExtDeviceMode call
	  // copy NewPort to TempNewPort to avoid corrupting profile string
	  lstrcpy(TempNewPort, NewPort);
	  (*lpfnExtDeviceMode)(hWindow, hPrintDrv, NULL,
				    NewName,
				    TempNewPort,
				    lpDevMode,
				    NULL, DM_MODIFY|DM_UPDATE);

	  lstrcpy(NewDriverNoExt, NewDriver);
          _fstrtok(NewDriverNoExt, ".");
	  // write a new setting for the default printer
	  wsprintf(NewPrinter, (LPSTR)"%s,%s,%s", NewName, NewDriverNoExt,
	           NewPort);
	  if (!WriteProfileString((LPSTR)"windows", (LPSTR)"device",
				  NewPrinter)) {
	     MsgIniWriteError(hWindow, IniFile);
	  }

	  // notify all windows of win.ini change
	  SendMessage(HWND_BROADCAST, WM_WININICHANGE, 0, (DWORD)"windows");

	  GlobalUnlock(hDevMode);
	  GlobalFree(hDevMode);

       }
       else {
	  MsgNoExtDeviceMode(hWindow);
       }
       FreeLibrary(hPrintDrv);
     }
     else {
       MsgNoLoad(hWindow);
     }
   }
   else {
     //  If profile string was set by SetupPrinter, it will be OK
     //  If user edits .ini, setting may be invalid
     MsgNotValidSetting(hWindow, pIniFile, pApplication);
     //  If any of required settings is invalid, must
     //  be sure NewName is a null string, since it is tested
     *NewName = '\0';
     
   }
   return 1;

}



/************************************
Reset NewDriver to its condition before last call to
SetPrinter and reset default printer profile string
************************************/

// Argument template = "Ay"

BOOL FAR PASCAL _export ResetPrinter(HANDLE hWindow)
{
   // reset the NewDriver
   ResetNewDriver(hWindow);

   // guard against user calling ResetPrinter first
   // do not eliminate existing device setting in such a case
   if (*NewName) {
      // reset the default printer profile string
      if (!WriteProfileString("windows", "device", OldPrinter)){
	 MsgIniWriteError(hWindow, IniFile);
      }

      // notify all windows
      SendMessage(HWND_BROADCAST, WM_WININICHANGE, 0, (DWORD)"windows");

      // set NewName to null string
      // so printer won't be reset until
      // after another valid call to SetPrinter 
      *NewName = '\0';
   }

   return 1;
}


/*****************************************
Print a form to printer specified in .ini file
*****************************************/
// Argument template = "ACCCny"

BOOL FAR PASCAL _export PrintFormX(LPSTR FormName,
				   LPSTR pIniFile,
				   LPSTR pApplication,
				   OVCALLBACK_EXECUTESTRING fnOVExec,
				   HANDLE hWindow)
{
   char  Buffer[80];

   SetPrinter(pIniFile, pApplication, hWindow);

   // print the OV Form
   wsprintf(Buffer, "@PRINTFORM(\"%s\")", FormName);
   (fnOVExec)(Buffer, NULL, 0, 1);

   ResetPrinter(hWindow);

   return 1;

}
